package hirondelle.web4jtools.codegenerator.feature;

import hirondelle.web4j.model.ModelCtorException;
import hirondelle.web4j.model.ModelUtil;
import hirondelle.web4jtools.util.Check;
import hirondelle.web4j.util.Util;
import static hirondelle.web4j.util.Consts.FAILS;
import hirondelle.web4jtools.codegenerator.codes.ApplyOperation;
import hirondelle.web4jtools.codegenerator.codes.ShowOperation;
import hirondelle.web4jtools.codegenerator.codes.UiStyle;

/**
* Model Object for your feature's basic information. 
* 
* <P>When generating code for a new feature, you start by entering this data. 
*/
public final class Feature {

  /**
  * Constructor.
  * 
  * <P>The feature name is used to generate names for many different items that form 
  * part of the implementation. For example, the base name of "Jet Engine" (with the space)
  * will be used to generate names such as :
  * <ul>
  * <li><tt>JetEngineAction.java</tt> - the Action class
  * <li><tt>JetEngine.java</tt> - the Model Object
  * <li><tt>fJetEngine</tt> - a field reference in the Action class
  * <li>and so on...
  * </ul>
  * 
  * <P>The formatting of the name depends on the context in which it is used, and is mostly 
  * controlled by the {@link hirondelle.web4jtools.util.Functions} utility class. You can implement 
  * custom coding conventions by altering the implementation of <tt>Functions</tt>.
  * 
  * <P>Warning : the text items in this class are modeled as <tt>String</tt>, not 
  * {@link hirondelle.web4j.security.SafeText}. If rendering this data in markup, it is 
  * strongly recommended that the text be escaped for special characters using <tt>w:safe()</tt>.
  * 
  * @param aName for the feature, 2..50 characters, required
  * @param aUiStyle denotes the style of user interaction, required
  * @param aPackageName name of the package for generated items, 2..200 characters, valid package name, required
  * @param aShowOperation either <tt>List</tt> or <tt>Fetch</tt>, optional. Used only with <tt>ActionTemplateShowAndApply</tt>.
  * @param aApplyOperation one of <tt>Add</tt>, <tt>Change</tt>, or <tt>Delete</tt>, optional. Used only with <tt>ActionTemplateShowAndApply</tt>.
  */
  public Feature(String aName, String aUiStyle, String aPackageName, String aShowOperation, String aApplyOperation) throws ModelCtorException {
    fName =  aName;
    fUiStyle = Util.textHasContent(aUiStyle) ? UiStyle.valueOf(aUiStyle) : null;
    fPackageName = aPackageName;
    fShowOperation = Util.textHasContent(aShowOperation) ? ShowOperation.valueOf(aShowOperation) : null;
    fApplyOperation = Util.textHasContent(aApplyOperation) ? ApplyOperation.valueOf(aApplyOperation) : null;
    validateState();
  }

  public String getName() { return fName; }
  public UiStyle getUiStyle() { return fUiStyle; }
  public String getPackageName() { return fPackageName; }
  public ShowOperation getShowOperation() { return fShowOperation; }
  public ApplyOperation getApplyOperation() { return fApplyOperation; }
  
  /** Intended for debugging only.  */
  @Override public String toString() {
    return ModelUtil.toStringFor(this);
  }

  @Override public boolean equals( Object aThat ) {
    Boolean result = ModelUtil.quickEquals(this, aThat);
    if ( result == null ){
      Feature that = (Feature) aThat;
      result = ModelUtil.equalsFor(this.getSignificantFields(), that.getSignificantFields());
    }
    return result;    
  }

  @Override public int hashCode() {
    if ( fHashCode == 0 ) {
      fHashCode = ModelUtil.hashCodeFor(getSignificantFields());
    }
    return fHashCode;
  }
  
  // PRIVATE //
  private final String fName;
  private final UiStyle fUiStyle;
  private final String fPackageName;
  private final ShowOperation fShowOperation;
  private final ApplyOperation fApplyOperation;
  private int fHashCode;
  
  private void validateState() throws ModelCtorException {
    ModelCtorException ex = new ModelCtorException();
    if ( FAILS == Check.required(fName, Check.range(2,50))) {
      ex.add("Feature Name is required, 2..50 chars.");
    }
    if ( FAILS == Check.required(fUiStyle) ) {
      ex.add("UI Style is required.");
    }
    if ( FAILS == Check.required(fPackageName, Check.range(2,200))) {
      ex.add("Package name is required, 2..200 characters.");
    }
    
    if ( fUiStyle == UiStyle.ListAndEdit ) {
      if ( fShowOperation != null ) {
        ex.add("Show Operation applies only if UI Style is Show and Apply");
      }
      if ( fApplyOperation != null ) {
        ex.add("Apply Operation applies only if UI Style is Show and Apply");
      }
    }
    
    if ( fUiStyle == UiStyle.ShowAndApply ){
      if ( fShowOperation == null ){
        ex.add("Please select a Show Operation.");
      }
      if ( fApplyOperation == null ){
        ex.add("Please select an Apply Operation.");
      }
    }
    
    if ( ! ex.isEmpty() ) throw ex;
  }
  
  private Object[] getSignificantFields(){
    return new Object[]{fName, fUiStyle, fPackageName, fShowOperation, fApplyOperation};
  }
}
