package hirondelle.web4jtools.codegenerator.field;

import hirondelle.web4j.model.ModelCtorException;
import hirondelle.web4j.model.ModelUtil;
import hirondelle.web4jtools.util.Check;
import hirondelle.web4j.util.Util;
import static hirondelle.web4j.util.Consts.FAILS;
import hirondelle.web4jtools.codegenerator.codes.ControlStyle;
import hirondelle.web4jtools.codegenerator.codes.FieldType;
import hirondelle.web4jtools.codegenerator.codes.SortOrder;
import hirondelle.web4j.model.Id;

/**
* A data field used by the feature.
* 
* <P>A data field usually maps to a database table column.
* 
* <P>One of these fields must be marked as the primary key.
*/
public final class Field {
  
  /**
  * Constructor. 
  *  
  * <P>Here, the <tt>Id</tt> is not a database key. Rather, it is an index into a collection of Fields 
  * stored only in memory. It acts much as a primary key, however.  
  * 
  * <P><b>'Required'  Requires Some Explanation</b><br> 
  * <span class='highlight'>Items are required in the sense of needing to be non-null in the Model Object. 
  * They are <em>not</em> required in the sense of being non-null in the database.</span> 
  * Usually this distinction is irrelevant. However, for primary keys the distinction is indeed relevant : primary keys are 
  * required in the database, but often <em>not</em> required in the Model Object. 
  * When adding a new record, a Model Object is first created, to model the user input. 
  * At this stage, however, the primary key is usually not yet defined. This is particularly true when the database 
  * autogenerates its primary keys when new records are added. 
  * 
  * <P><b>Minimum and Maximum</b><br>
  * These are numeric entries, and are interpreted differently according to {@link FieldType}. For numeric data, they are simply 
  * limits on the field's value. For text data, they are limits on the number of characters. 
  * See {@link hirondelle.web4j.model.Check} for more information on its <tt>min</tt> and <tt>max</tt> methods.
  * 
  * <P><b>Hard Validation</b><br>
  * One and only one form of 
  * <a href='http://www.web4j.com/web4j/javadoc/hirondelle/web4j/security/ApplicationFirewall.html'>hard validation</a> 
  * must be selected, either <em>length</em> or <em>pattern</em>, but not both.
  * 
  * <P><b>Soft Validation</b><br>
  * See 
  * <a href='http://www.web4j.com/web4j/javadoc/hirondelle/web4j/security/ApplicationFirewall.html'>ApplicationFirewall</a>
  * and 
  * <a href='http://www.web4j.com/web4j/javadoc/hirondelle/web4j/model/Check.html'>Check</a> for 
  * more information on soft validation.
  * 
  * @param aId optional, since when adding new records the id is unknown. 
  * @param aName is required, 1..100 characters; enter as natural text, such as <tt>'Jet Engine'</tt>, with a space.
  * @param aDescription is optional, 1..1000 characters.
  * @param aIsRequired is optional; required fields are tested for non-nullity in the Model Object. 
  * @param aType is required, defines the {@link FieldType}, the java class used to represent the field 
  * @param aControlStyle is required, defines the {@link ControlStyle}, the HTML control used to enter the field value.
  * @param aIsPrimaryKey is optional; only one field in the feature should be the primary key.
  * @param aMinimum is optional, cannot be greater than <tt>aMaximum</tt>.
  * @param aMaximum is optional, cannot be less than <tt>aMinimum</tt>.
  * @param aNumDecimals is optional, integer, <tt>1</tt> or more.
  * @param aErrorMessage is optional, the error message displayed when the user inputs an invalid value, 1..200 characters
  * @param aIsHardValidatedForLength is optional.
  * @param aHardValidationPattern is optional, regular expression, 1..200 characters.
  * @param aCheckPattern is optional, regular expression, 1..200 characters. 
  * @param aCheckEmail is optional, <tt>true</tt> only if this field contains an email address.
  * @param aCheckSpam is optional, <tt>true</tt> only if this is a text field that should be checked for spam.
  * @param aIsOrderByField is optional, <tt>true</tt> only if this is the field by which listings should be sorted; only one 
  * field in the feature should be the ORDER BY field.
  * @param aIsDescendingOrder is optional, and takes effect only if this is the ORDER BY field.
  * @param aCodeTable is optional, and is the name of an application-scope code table; 
  * applies only if <tt>aControlStyle</tt> is <tt>SELECT</tt> or <tt>RADIO</tt>; 
  */
  public Field(
    Id aId, String aName, String aDescription, Boolean aIsRequired, String aType, String aControlStyle, Boolean aIsPrimaryKey, 
    Integer aMinimum, Integer aMaximum, Integer aNumDecimals, String aErrorMessage, Boolean aIsHardValidatedForLength, 
    String aHardValidationPattern, String aCheckPattern, Boolean aCheckEmail, Boolean aCheckSpam,
    Boolean aIsOrderByField, Boolean aIsDescendingOrder, String aCodeTable
  ) throws ModelCtorException {
    fId = aId;
    fName = aName;
    fDescription = aDescription;
    fIsRequired = Util.nullMeansFalse(aIsRequired);
    fType = Util.textHasContent(aType) ? FieldType.valueOf(aType) : null;
    fControlStyle = Util.textHasContent(aControlStyle) ? ControlStyle.valueOf(aControlStyle) : null;
    fIsPrimaryKey = Util.nullMeansFalse(aIsPrimaryKey);
    fMinimum = aMinimum;
    fMaximum = aMaximum;
    fNumDecimals = aNumDecimals;
    fErrorMessage = aErrorMessage;
    fIsHardValidatedForLength = Util.nullMeansFalse(aIsHardValidatedForLength);
    fHardValidationPattern = aHardValidationPattern;
    fCheckPattern = aCheckPattern;
    fCheckEmail = Util.nullMeansFalse(aCheckEmail);
    fCheckSpam = Util.nullMeansFalse(aCheckSpam);
    fIsOrderByField = Util.nullMeansFalse(aIsOrderByField);
    fIsDescendingOrder = Util.nullMeansFalse(aIsDescendingOrder);
    fCodeTable = aCodeTable;
    validateState();
  }

  public Id getId() { return fId; }
  public String getName() {  return fName;  }
  public String getDescription() { return fDescription; }
  public Boolean getIsRequired() { return fIsRequired;  }
  public FieldType getType() { return fType;  }
  public ControlStyle getControlStyle() { return fControlStyle; } 
  public Boolean getIsPrimaryKey() { return fIsPrimaryKey;  }
  public Integer getMinimum() { return fMinimum; }
  public Integer getMaximum() {  return fMaximum; }
  public Integer getNumDecimals(){ return fNumDecimals; }
  public String getErrorMessage() {  return fErrorMessage; }
  public Boolean getIsHardValidatedForLength() { return fIsHardValidatedForLength;  }
  public String getHardValidationPattern() { return fHardValidationPattern; }
  public String getCheckPattern() { return fCheckPattern; }
  public Boolean getCheckEmail() { return fCheckEmail; }
  public Boolean getCheckSpam() { return fCheckSpam; }
  public Boolean getIsOrderByField() { return fIsOrderByField; }
  public Boolean getIsDescendingOrder() { 
    return fIsOrderByField == null ? null : fIsDescendingOrder; 
  }
  public String getCodeTable(){ return fCodeTable; }
  
  /**
  * Return the {@link SortOrder} calculated from {@link #getIsOrderByField()} and {@link #getIsDescendingOrder()}.
  */
  public SortOrder getSortOrder() {
    SortOrder result = null;
    if( fIsOrderByField == true ) {
      result =  fIsDescendingOrder ? SortOrder.DESC : SortOrder.ASC; 
    }
    return result;
  }
  
  /** Intended for debugging only.  */
  @Override public String toString() {
    return ModelUtil.toStringFor(this);
  }

  @Override public boolean equals( Object aThat ) {
    Boolean result = ModelUtil.quickEquals(this, aThat);
    if ( result == null ){
      Field that = (Field) aThat;
      result = ModelUtil.equalsFor(this.getSignificantFields(), that.getSignificantFields());
    }
    return result;    
  }

  @Override public int hashCode() {
    if ( fHashCode == 0 ) {
      fHashCode = ModelUtil.hashCodeFor(getSignificantFields());
    }
    return fHashCode;
  }
  
  // PRIVATE //
  private final Id fId;
  private final String fName;
  private final String fDescription;
  private final Boolean fIsRequired;
  private final FieldType fType;
  private final ControlStyle fControlStyle;
  private final Boolean fIsPrimaryKey;
  private final Integer fMinimum;
  private final Integer fMaximum;
  private final Integer fNumDecimals;
  private final String fErrorMessage;
  private final Boolean fIsHardValidatedForLength;
  private final String fHardValidationPattern;
  private final String fCheckPattern;
  private final Boolean fCheckEmail;
  private final Boolean fCheckSpam;
  private final Boolean fIsOrderByField;
  private final Boolean fIsDescendingOrder;
  private final String fCodeTable;
  
  private int fHashCode;
  
  private void validateState() throws ModelCtorException {
    ModelCtorException ex = new ModelCtorException();
    if ( FAILS == Check.required(fName, Check.range(1,100)) ) {
      ex.add("Name required, 1..100 characters.");
    }
    if ( FAILS == Check.required(fType) ) {
      ex.add("Type is required.");
    }
    if ( FAILS == Check.required(fControlStyle) ) {
      ex.add("Control Style is required.");
    }
    if ( FAILS == Check.optional(fDescription, Check.range(1,1000))) {
      ex.add("Description is optional, 1..1000 characters.");
    }
    if ( FAILS == Check.optional(fErrorMessage, Check.range(1,200)) ) {
      ex.add("Message for invalid input is optional, 1..200 characters.");
    }
    if ( FAILS == Check.optional(fHardValidationPattern, Check.range(1,200), Check.isRegex()) ) {
      ex.add("Hard-validation pattern optional, regular expression, 1..200 characters.");
    }
    if ( fIsHardValidatedForLength == true && Util.textHasContent(fHardValidationPattern)) {
      ex.add("Hard validation for length and for pattern are mutually exclusive. Please select one or the other.");
    }
    if ( fIsHardValidatedForLength == false && ! Util.textHasContent(fHardValidationPattern)) {
      ex.add("Please select a style of hard validation.");
    }
    if( FAILS == Check.optional(fCheckPattern, Check.isRegex(), Check.range(1,200))) {
      ex.add("Soft Validation pattern is not a regular expression, 1..200 characters.");
    }
    if (fIsPrimaryKey == true && fIsRequired == true ) {
      ex.add("For Model Objects, primary key cannot be required: 'Add' operations need an empty primary key.");
    }
    if ( fMinimum != null && fMaximum != null ) {
      if( fMinimum.compareTo(fMaximum) > 0 ) {
        ex.add("Minimum cannot be greater than maximum");
      }
    }
    if ( FAILS == Check.optional(fNumDecimals, Check.min(1))) {
      ex.add("Number of decimals must be 1 or more.");
    }
    if( needsCodeTable(fControlStyle) && ! Util.textHasContent(fCodeTable)) {
      ex.add("Please enter a code table.");
    }
    if ( Util.textHasContent(fCodeTable) && ! needsCodeTable(fControlStyle)) {
      ex.add("Please remove code table. Code tables are not used with this kind of control.");
    }
    if ( ex.isNotEmpty() ) { throw ex; }
  }
  
  private boolean needsCodeTable(ControlStyle aStyle){
    return  (aStyle == ControlStyle.Select || aStyle == ControlStyle.Radio);
  }
  
  private Object[] getSignificantFields(){
    return new Object[]{
      fName, fDescription, fIsRequired, fType, fControlStyle, fIsPrimaryKey, fMinimum, fMaximum, fNumDecimals, 
      fErrorMessage, fIsHardValidatedForLength, fHardValidationPattern, fIsOrderByField, fIsDescendingOrder,
      fCodeTable
    };
  }
}
