package hirondelle.web4jtools.logview.directories;

import hirondelle.web4j.model.ModelCtorException;
import hirondelle.web4j.security.SafeText;
import hirondelle.web4j.model.Check;
import hirondelle.web4j.model.ModelUtil;
import hirondelle.web4j.model.Validator;
import java.io.*;
import static hirondelle.web4j.util.Consts.FAILS;
import static hirondelle.web4j.util.Consts.FILE_SEPARATOR;
import hirondelle.web4j.util.Util;

/**
* Basic information on logging directories. 
* 
* <P>This tool can examine both your applictation logs, and the logs of your 
* server (such as Tomcat). 
*/
public final class LogInfo {
  
  /**
  * Full constructor.
  * 
  * @param aProjectName simple name or alias for the project, displayed as a reminder (required).
  * @param aAppLogDir directory containing application log files (required); existing directory, ending in a separator; 
  * any files having no size (such as <tt>.lck</tt> files) will be ignored, while all other files in this directory will be considered log files.
  * @param aServerLogDir directory containing your server's log files (required); existing directory, ending in a separator.
  * @param aServerLogStartsWith server log files of interest have file names starting with this text (for example, <tt>'localhost'</tt>); 
  * for Tomcat, several log files are generated, but it is almost always the case that only one kind of log file has items of 
  * interest.
  */ 
  public LogInfo(SafeText aProjectName, SafeText aAppLogDir, SafeText aServerLogDir, SafeText aServerLogStartsWith) throws ModelCtorException {
    fProjecName = aProjectName;
    fAppLogDir = aAppLogDir;
    fServerLogDir = aServerLogDir;
    fServerLogStartsWith = aServerLogStartsWith;
    validateState();
  }
  
  public SafeText getAppLoggingDirectory() {  return fAppLogDir; }
  public SafeText getProjectName() {  return fProjecName; }
  public SafeText getServerLoggingDirectory() {  return fServerLogDir;  }
  public SafeText getServerLogFileStartsWith() {  return fServerLogStartsWith;  }
  
  /** Intended for debugging only.  */
  @Override public String toString() {
    return ModelUtil.toStringFor(this);
  }

  @Override public boolean equals( Object aThat ) {
    Boolean result = ModelUtil.quickEquals(this, aThat);
    if ( result == null ){
      LogInfo that = (LogInfo) aThat;
      result = ModelUtil.equalsFor(this.getSignificantFields(), that.getSignificantFields());
    }
    return result;    
  }

  @Override public int hashCode() {
    if ( fHashCode == 0 ) {
      fHashCode = ModelUtil.hashCodeFor(getSignificantFields());
    }
    return fHashCode;
  }
  
  // PRIVATE //
  private SafeText fProjecName;
  private SafeText fAppLogDir;
  private SafeText fServerLogDir;
  private SafeText fServerLogStartsWith;
  private int fHashCode;
  
  private void validateState() throws ModelCtorException {
    ModelCtorException ex = new ModelCtorException();
    if( FAILS == Check.required(fProjecName) ) {
      ex.add("Project Name is required.");
    }
    if ( FAILS == Check.required(fAppLogDir, validDirectory())) {
      ex.add("Application Logging Directory is required: an existing, read-able directory, ending with a " + Util.quote(FILE_SEPARATOR) + ".");
    }
    if ( FAILS == Check.required(fServerLogDir, validDirectory())) {
      ex.add("Tomcat Logging Directory is required: an existing, read-able directory, ending with a " + Util.quote(FILE_SEPARATOR) + ".");
    }
    if ( FAILS == Check.required(fServerLogStartsWith)) {
      ex.add("Tomcat Log File Starts With is required.");
    }
    if ( ex.isNotEmpty() ) throw ex;
  }
  
  private Validator validDirectory() {
    return new Validator() {
      public boolean isValid(Object aObject) {
        SafeText text = (SafeText)aObject;
        File file = new File(text.getRawString());
        return file.isDirectory() && file.exists() && file.canRead() && file.isAbsolute() && endsWithSeparator(text);
      }
    };
  }
  
  private boolean endsWithSeparator(SafeText aDir){
    return aDir.getRawString().endsWith(FILE_SEPARATOR);
  }
  
  private Object[] getSignificantFields(){
    return new Object[] {fProjecName, fAppLogDir, fServerLogDir, fServerLogStartsWith};
  }
}
