package hirondelle.web4jtools.logview.downtime;

import hirondelle.web4j.model.Check;
import hirondelle.web4j.model.ModelCtorException;
import hirondelle.web4j.model.ModelUtil;
import static hirondelle.web4j.util.Consts.FAILS;
import java.util.Date;

/** 
* Model Object for application down time. 
* 
* <P>Here, 'down time' has a specific meaning. It is extracted from your <em>application</em> logs only, 
* and not from server logs. It is calculated by examining all the log files in your application's logging directory. 
* Both the last-modified date of the files themselves, and the date attached to the <em>first</em>
* log record in each log file are examined. 
* 
* <P>This mechanism assumes that a new log file is created when the application restarts. 
* 
* <P>If A and B are two consecutive log files, then a down time is calculated as : <br> 
* <tt>(time of first log record in B) - (last-modified date for A)</tt>
* 
* <P>This class is immutable, and makes defensive copies.
*/
public final class DownTime {

  /**
  * Full constructor.
  *  
  * @param aFrom Start of down time interval (required).
  * @param aTo End of the down time interval (required).
  */
  public DownTime(Date aFrom, Date aTo) throws ModelCtorException {
    //defensive copies of mutable fields.
    fFrom = aFrom.getTime();
    fTo = aTo.getTime();
    validateState();
  }
  
  public Date getFrom() { return new Date(fFrom); }  
  public Date getTo() { return new Date(fTo); }
  
  /** Duration of the down time, in minutes.  */
  public Long getDuration() {
    long result = 0;
    if ( fTo >= fFrom ) {
      result = (fTo - fFrom)/(1000*60); 
    }
    return result;
  }
  
  @Override public String toString(){
    return ModelUtil.toStringFor(this);
  }
  
  @Override public boolean equals(Object aThat){
    Boolean result = ModelUtil.quickEquals(this, aThat);
    if ( result == null ) {
      DownTime that = (DownTime)aThat;
      result = ModelUtil.equalsFor(this.getSignificantFields(), that.getSignificantFields());
    }
    return result;
  }
  
  @Override public int hashCode(){
    if ( fHashCode == 0 ) {
      fHashCode = ModelUtil.hashCodeFor(getSignificantFields());
    }
    return fHashCode;
  }

  // PRIVATE //
  private final Long fFrom;
  private final Long fTo;
  private int fHashCode;

  private void validateState() throws ModelCtorException {
    ModelCtorException ex = new ModelCtorException();
    
    if ( FAILS == Check.required(fFrom) ) {
      ex.add("From is required.");
    }
    if ( FAILS == Check.optional(fTo) ) {
      ex.add("To is required.");
    }

    if ( ! ex.isEmpty() ) throw ex;
  }
  
  private Object[] getSignificantFields(){
    return new Object[] {fFrom, fTo};
  }
}