package hirondelle.web4jtools.logview.parsedview;

import hirondelle.web4j.model.Check;
import hirondelle.web4j.model.ModelCtorException;
import hirondelle.web4j.model.ModelUtil;
import static hirondelle.web4j.util.Consts.FAILS;
import static hirondelle.web4j.util.Consts.NOT_FOUND;
import hirondelle.web4j.security.SafeText;
import java.util.Date;
import hirondelle.web4j.util.Util;
import hirondelle.web4jtools.logview.simpleview.LogFor;
import java.util.logging.Level;

/** 
*  Model Object for the criteria used for viewing parsed log records.
*  
* <P>This class is immutable, and makes defensive copies.
*/
public final class ParsedCriteria {

  /**
  * Full constructor.
  *
  * @param aLogFor either application log or server log (required).
  * @param aLevel minimum JDK Logger Level (required); any record at this level or higher will be displayed.
  * @param aMinDate minimum Date for log message (cannot come after <tt>aMaxDate</tt>).
  * @param aMaxDate maximum Date for log messages.
  * @param aLoggerStartsWith logger/package name starts with this text.
  * @param aMethodName name of the method emitting the log message.
  * @param aLogMessageContains log message must contain this text.
  * @param aReverseOrder doesn't filter records, but rather sorts them in reverse order, placing most recent first.
  */
  public ParsedCriteria(
   SafeText aLogFor,
   SafeText aLevel, 
   Date aMinDate, 
   Date aMaxDate, 
   SafeText aLoggerStartsWith, 
   SafeText aMethodName,
   SafeText aLogMessageContains,
   Boolean aReverseOrder
  ) throws ModelCtorException {
    fLogFor = Util.textHasContent(aLogFor) ? LogFor.valueOf(aLogFor.getRawString()) : null;
    fLevel = Util.textHasContent(aLevel) ? Level.parse(aLevel.getRawString()) : null;
    fMinDate = aMinDate == null ? null : aMinDate.getTime();
    fMaxDate = aMaxDate == null ? null : aMaxDate.getTime();
    fLoggerStartsWith = aLoggerStartsWith;
    fMethodName = aMethodName;
    fLogMessageContains = aLogMessageContains;
    fReverseOrder = Util.nullMeansFalse(aReverseOrder);
    validateState();
  }
  
  /** 
  * Return an instance having <tt>FINEST</tt> level for application logs.
  * (No other criteria.)
  */
  public static ParsedCriteria forDowntimeListing() {
    try {
      return new ParsedCriteria(SafeText.from(LogFor.Application.name()), SafeText.from(Level.FINEST.getName()), null, null, null, null, null, null);
    }
    catch(ModelCtorException ex){
      throw new RuntimeException(ex);
    }
  }
  
  /** 
  * Return an instance having <tt>FINEST</tt> level for the given log.
  * (No other criteria.)
  */
  public static ParsedCriteria forStats(String aLogFor) throws ModelCtorException {
    return new ParsedCriteria(SafeText.from(aLogFor), SafeText.from(Level.FINEST.getName()), null, null, null, null, null, null);
  }
  
  public LogFor getLogFor() { return fLogFor; }
  Level getLevel() { return fLevel; }  
  Date getMinDate() { return new Date(fMinDate); }  
  Date getMaxDate() { return new Date(fMaxDate); }  
  SafeText getLoggerStartsWith() { return fLoggerStartsWith; }
  SafeText getMethodName() { return fMethodName; }
  SafeText getLogMessageContains() { return fLogMessageContains; }
  public Boolean getReverseOrder() { return fReverseOrder; }
  
  /**
  * Return <tt>true</tt> only if the given logger record passes these criteria.
  */
  public boolean passes(LoggerRecord aRecord) {
    boolean result = true;
    if( aRecord.getLevel().intValue() < fLevel.intValue() ) {
      result = FAILS; 
    }
    if( fMinDate != null  ) {
      if ( aRecord.getDate().getTime() < fMinDate ) {
        result = FAILS;
      }
    }
    if( fMaxDate != null  ) {
      if ( aRecord.getDate().getTime() > fMaxDate ) {
        result = FAILS;
      }
    }
    if ( Util.textHasContent(fLoggerStartsWith) ) {
      if ( ! aRecord.getLogger().getRawString().startsWith(fLoggerStartsWith.getRawString()) ) {
        result = FAILS;
      }
    }
    if ( Util.textHasContent(fLogMessageContains) ) {
      if ( NOT_FOUND == aRecord.getMessage().getRawString().indexOf(fLogMessageContains.getRawString()) ) {
        result = FAILS;
      }
    }
    if ( Util.textHasContent(fMethodName) ) {
      if ( ! aRecord.getMethod().getRawString().equalsIgnoreCase(fMethodName.getRawString()) ) {
        result = FAILS;
      }
    }
    return result;
  }
  
  @Override public String toString(){
    return ModelUtil.toStringFor(this);
  }
  
  @Override public boolean equals(Object aThat){
    Boolean result = ModelUtil.quickEquals(this, aThat);
    if ( result == null ) {
      ParsedCriteria that = (ParsedCriteria)aThat;
      result = ModelUtil.equalsFor(this.getSignificantFields(), that.getSignificantFields());
    }
    return result;
  }
  
  @Override public int hashCode(){
    if ( fHashCode == 0 ) {
      fHashCode = ModelUtil.hashCodeFor(getSignificantFields());
    }
    return fHashCode;
  }

  // PRIVATE //
  private final LogFor fLogFor;
  private final Level fLevel;
  private final Long fMinDate;
  private final Long fMaxDate;
  private final SafeText fLoggerStartsWith;
  private final SafeText fMethodName;
  private final SafeText fLogMessageContains;
  private final Boolean fReverseOrder;
  private int fHashCode;

  private void validateState() throws ModelCtorException {
    ModelCtorException ex = new ModelCtorException();
    
    if( FAILS == Check.required(fLogFor) ) {
      ex.add("Log For is required.");
    }
    if ( FAILS == Check.required(fLevel) ) {
      ex.add("Level is required.");
    }
    if ( fMinDate != null && fMaxDate != null  ) {
      if( fMinDate > fMaxDate ) {
        ex.add("Min Date/Time must come before Max Date/Time");
      }
    }

    if ( ! ex.isEmpty() ) throw ex;
  }
  
  private Object[] getSignificantFields(){
    return new Object[] {fLogFor, fLevel, fMinDate, fMaxDate, fLoggerStartsWith, fMethodName, fLogMessageContains, fReverseOrder};
  }
}