package hirondelle.web4jtools.logview.parser;

import java.util.*;
import hirondelle.web4j.util.Util;
import java.util.logging.*;
import static hirondelle.web4j.util.Consts.NEW_LINE;
import static hirondelle.web4j.util.Consts.SPACE;
import static hirondelle.web4j.util.Consts.NOT_FOUND;
import java.text.SimpleDateFormat;

import hirondelle.web4jtools.logview.parsedview.LoggerRecord;
import hirondelle.web4jtools.logview.parsedview.ParsedCriteria;

/**
* Parse a JDK log file having default format.
*
* <P>Example of a typical log record :
* <PRE>
16-Sep-2011 9:00:28 PM hirondelle.web4j.webmaster.LoggingConfigImpl tryTestMessages
FINE: This is a test message for Logger 'hirondelle.web4jtools'
</PRE>
*
* The above record is parsed by this class into the following parts :
*<ul>
* <li>Date - <tt>16-Sep-2011 9:00:28 PM</tt>
* <li>Logger - <tt>hirondelle.web4j.webmaster.LoggingConfigImpl</tt>
* <li>Method - <tt>tryTestMessages</tt>
* <li>Level - <tt>FINE</tt>
* <li>Message - <tt>This is a test message for Logger 'hirondelle.web4jtools'</tt>
*</ul>
*
* <P>The message is often over multiple lines.
* 
* <P>There is no specific delimiter for this format, so this implementation uses an 
* <em>ad hoc</em> method of delimiting records.
* 
* <P>Note that this class is package-private, to prevent it from being used directly by 
* other packages. The caller uses 
* {@link hirondelle.web4jtools.logview.parser.LogParserInstance} to obtain instances of  
* this class. 
*/
final class LogParserForJDKDefault implements LogParser {

  /** The assumed date format for logging output - {@value}. You may need to change this to reflect your host's settings. */
  public static final String DATE_FORMAT = "d-MMM-yyyy h:mm:ss a";

  public List<LoggerRecord> parse(String aLogFileContents, ParsedCriteria aCriteria) {
    List<LoggerRecord> result = new ArrayList<LoggerRecord>();
    LoggerRecord currentLoggerRecord = null;
    StringBuilder currentMsg = new StringBuilder(); //starts with 2nd line content; may have more
    Scanner scanner = new Scanner(aLogFileContents);
    while (scanner.hasNextLine()) {
      String line = scanner.nextLine() + NEW_LINE;
      if ( isFirstLine(line) ) {
        fLogger.fine("FIRST: "+ Util.quote(line));
        finishOldLoggerRecord(aCriteria, result, currentLoggerRecord, currentMsg);
        currentLoggerRecord = startNewLoggerRecord(line);
      }
      else if ( isSecondLine(line) ) {
        fLogger.fine("SECOND: "+ Util.quote(line));
        currentLoggerRecord.addLevel( extractLevelFromSecond(line) );
        currentMsg = new StringBuilder( extractMessageFromSecond(line) );
      }
      else {
        fLogger.fine("OTHER: "+ Util.quote(line));
        currentMsg.append(line);
      }
    }
    finishOldLoggerRecord(aCriteria, result, currentLoggerRecord, currentMsg);
    scanner.close();
    return result;
  }

  // PRIVATE 
  private static final Logger fLogger = Util.getLogger(LogParserForJDKDefault.class);
  
  private boolean isFirstLine(String aLine){
    return startsWithDateTime(aLine);
  }
  
  private boolean startsWithDateTime(String aLine){
    //not fixed width. Java date parsing in weird. Accepts trailing chars.
    //16-Sep-2011 10:32:25 PM hirondelle.web4j...
    //6-Sep-2011 7:30:53 PM org.apache.cata....
    boolean result = false;
    if(aLine.length()>22){
      String date = aLine.substring(0,23).trim();
      SimpleDateFormat dateFormat = new SimpleDateFormat(DATE_FORMAT);
      try {
        Date testDate = dateFormat.parse(date);
        result = true;
      }
      catch(java.text.ParseException ex) {
        //do nothing at all
      }
    }
    return result;
  }
  
  private boolean isSecondLine(String aLine){
    return 
      aLine.startsWith("SEVERE") ||
      aLine.startsWith("WARNING") ||
      aLine.startsWith("INFO") ||
      aLine.startsWith("CONFIG") ||
      aLine.startsWith("FINE") ||
      aLine.startsWith("FINER") ||
      aLine.startsWith("FINEST")
    ;
  }
  
  private LoggerRecord startNewLoggerRecord(String aLine){
    //6-Sep-2011 10:55:09 AM hirondelle.web4j.Controller init
    int idxMethod = aLine.lastIndexOf(SPACE);
    if(idxMethod == NOT_FOUND) {
      fLogger.severe("Cannot find Method in " + Util.quote(aLine));
    }
    String method = aLine.substring(idxMethod+1);
    
    int idxLogger = aLine.lastIndexOf(SPACE, idxMethod-1);
    if(idxLogger == NOT_FOUND) {
      fLogger.severe("Cannot find Logger in " + Util.quote(aLine));
    }
    String logger = aLine.substring(idxLogger + 1, idxMethod);
    
    Date date = parseDate(aLine.substring(0,idxLogger));
    return new LoggerRecord(date, logger, method);
  }
  
  private Date parseDate(String aDate) {
    //Since the source date is generated by code, it is less dangerous to use SimpleDateFormat.
    //6-Sep-2011 10:55:09 AM
    Date result = null;
    SimpleDateFormat dateFormat = new SimpleDateFormat(DATE_FORMAT);
    try {
      result = dateFormat.parse(aDate);
    }
    catch(java.text.ParseException ex) {
      fLogger.severe("Cannot parse into a date : " + aDate + " using format " + Util.quote(DATE_FORMAT));
    }
    return result;
  }

  private String extractLevelFromSecond(String aLine){
    int idxColon = aLine.indexOf(":");
    return aLine.substring(0,idxColon);
  }

  private String extractMessageFromSecond(String aLine) {
    int idxColon = aLine.indexOf(":");
    return aLine.substring(idxColon+1);
  }
  
  private void finishOldLoggerRecord(ParsedCriteria aCriteria, List<LoggerRecord> aResult, LoggerRecord aCurrentLoggerRecord, StringBuilder aCurrentMsg) {
    if( aCurrentLoggerRecord != null ) {
      aCurrentLoggerRecord.addMessage(aCurrentMsg.toString());
      if( aCriteria.passes(aCurrentLoggerRecord)) {
        aResult.add(aCurrentLoggerRecord);
      }
    }
  } 
}
