package hirondelle.web4jtools.logview.simpleview;

import static hirondelle.web4j.util.Consts.NEW_LINE;
import hirondelle.web4j.database.DAOException;
import hirondelle.web4j.util.Util;
import hirondelle.web4jtools.logview.directories.LogInfo;
import hirondelle.web4jtools.logview.directories.LogInfoDAO;
import hirondelle.web4jtools.logview.parsedview.LoggerRecord;
import hirondelle.web4jtools.logview.parsedview.ParsedCriteria;
import hirondelle.web4jtools.util.Ensure;
import hirondelle.web4jtools.logview.parser.LogParser;
import hirondelle.web4jtools.logview.parser.LogParserInstance;

import java.io.File;
import java.io.FileFilter;
import java.io.FileNotFoundException;
import java.util.*;
import java.util.logging.Logger;

import javax.servlet.ServletConfig;
import javax.servlet.http.HttpServletRequest;

/**
* Data Access Object for log files.
* 
*<P>This class is used both for viewing log files as plain text, and for 
* parsing log files into {@link hirondelle.web4jtools.logview.parsedview.LoggerRecord}s.
*/
public final class LogFileDAO {

  /** Read  in config from <tt>web.xml</tt>. */
  public static void readConfig(ServletConfig aConfig){
    fAPP_ENCODING = aConfig.getInitParameter(APP_ENCODING);
    fSERVER_ENCODING = aConfig.getInitParameter(SERVER_ENCODING);
    Ensure.isPresentInWebXml(APP_ENCODING, fAPP_ENCODING);
    Ensure.isPresentInWebXml(SERVER_ENCODING, fSERVER_ENCODING);
  }
  
  /** Full constructor. */
  public LogFileDAO(HttpServletRequest aRequest){
    fDirInfoDAO = new LogInfoDAO(aRequest);  
  }
  
  /**
  * Return the most recent log file in a given directory.
  *   
  * Returns <tt>null</tt> if no file found in the logging directory.  
  */
  public File getMostRecentLogFile(LogFor aLogFor) throws DAOException {
    return getMostRecentFile(aLogFor);
  }
  
  /**
  * Return the contents of a log file as a <tt>String</tt>.
  * 
  * <P>If the given criteria desire only a section of the file, then only the first or last 
  * section is returned; otherwise, the contents of the entire file is returned.
  */
  public String getLogFileContents(File aFile, SimpleCriteria aCriteria) throws DAOException {
    String result = null;
    if( aCriteria.getSection() == null || aCriteria.getSection() == Section.First) {
      result = getAllOrFirst(aFile, aCriteria.getNumLines());
    }
    else {
      result = getEnd(aFile, aCriteria.getNumLines());
    }
    return result;
  }
  
  /**
  * Return the entire content of a log file as a {@code List<LoggerRecord>}.
  */
  public List<LoggerRecord> getParsedLogFile(File aFile, ParsedCriteria aCriteria) throws DAOException {
    LogParser parser = LogParserInstance.getFor(aCriteria.getLogFor());
    List<LoggerRecord> result = parser.parse(getAllOrFirst(aFile, null), aCriteria);
    if( aCriteria.getReverseOrder() ) {
      Collections.reverse(result);
    }
    return result;
  }
  
  /**
  * Return a <tt>List</tt> of all <em>application</em> log files.
  * 
  * Any files having zero size (such as lock files) are ignored. The files are returned in order of increasing 
  * 'last-modified' date, such that the oldest log is first. This method is used when calculating down-times 
  * from logs. 
  */
  public List<File> listAllAppLogFiles(){
    File appLogDir = getDirectory(LogFor.Application);
    File[] logFiles = appLogDir.listFiles(getFilter(LogFor.Application));
    List<File> result = Arrays.asList(logFiles); //assumes all are logs, with no directories underneath
    Collections.sort(result, byModifiedDateDesc());
    Collections.reverse(result);
    fLogger.fine("Number of app log files in app log directory : " + result.size());
    return result;
  }

  /**
  * Return a <tt>List</tt> containing the <em>first</em> {@link LoggerRecord} in each 
  * of the given log files. This method is used when calculating down times from logs.
  */
  public List<LoggerRecord> listFirstRecordsFor(List<File> aAllLogFiles) throws DAOException {
    List<LoggerRecord> result = new ArrayList<LoggerRecord>();
    for(File file : aAllLogFiles){
      result.add(firstRecordFor(file));
    }
    return result;
  }
  
  // PRIVATE //
  
  private final LogInfoDAO fDirInfoDAO;
  
  private static String SERVER_ENCODING = "ServerLogFileEncoding";
  private static String fSERVER_ENCODING;
  
  private static String APP_ENCODING = "ApplicationLogFileEncoding";
  private static String fAPP_ENCODING;
  
  private static final Logger fLogger = Util.getLogger(LogFileDAO.class);

  /** Returns null if no file in the directory at all.  */
  private File getMostRecentFile(LogFor aLogFor) {
    File result = null;
    List<File> allFiles = Arrays.asList(getDirectory(aLogFor).listFiles(getFilter(aLogFor)));
    Collections.sort( allFiles, byModifiedDateDesc() );
    if ( ! allFiles.isEmpty() ) {
      result =  allFiles.get(0);
      fLogger.fine("Most recent " + aLogFor + " log file : " + Util.quote(result.getAbsolutePath()));
    }
    else {
      fLogger.fine("No log files detected in directory : " + getDirectory(aLogFor));
    }
    return result;
  }
  
  private File getDirectory(LogFor aLogFor){
    File result = null;
    if( LogFor.Server == aLogFor ) {
      result = new File(fDirInfoDAO.fetch().getServerLoggingDirectory().getRawString());
    }
    else if ( LogFor.Application == aLogFor ) {
      result = new File(fDirInfoDAO.fetch().getAppLoggingDirectory().getRawString());
    }
    else {
      throw new AssertionError("Unknown value of LogFor enumeration.");
    }
    return result;
  }

  /** Sort files by date.  */
  private Comparator<File> byModifiedDateDesc(){
    return new Comparator<File>() {
      public int compare(File aThis, File aThat) {
        Long thisDate = aThis.lastModified();
        Long thatDate = aThat.lastModified();
        return thisDate.compareTo(thatDate) * (-1);
      }
    };
  }
  
  /** 
  * Return either all or a portion of a file as a String.
  * 
  * @param aNumLines num lines to return; if <tt>null</tt>, then all lines are returned.  
  */ 
  private String getAllOrFirst(File aFile, Integer aNumLines) throws DAOException {
    fLogger.fine("Getting either all of log file, or first section.");
    StringBuilder result = new StringBuilder();
    Scanner scanner = null;
    try {
      scanner = new Scanner(aFile);
    }
    catch(FileNotFoundException ex){
      throw new DAOException("Cannot find file " + Util.quote(aFile.getAbsolutePath()), ex);
    }
    int lineCount = 0;
    while ( scanner.hasNextLine()) {
      result.append(scanner.nextLine() + NEW_LINE);
      lineCount++;
      if ( aNumLines != null && lineCount == aNumLines) break;
    }
    scanner.close();
    fLogger.fine("Number of log lines returned : " + lineCount);
    return result.toString();
  }

  /**
  * Return the end section of a file as a String.
  * 
  * @param aNumLinesForDisplay number of lines to return.
  */
  private String getEnd(File aFile, Integer aNumLinesForDisplay) throws DAOException {
    fLogger.fine("Getting up to " + aNumLinesForDisplay + " lines from the end of the log file.");
    String result = null;
    Integer numLinesInFile = getNumLines(aFile);
    if ( numLinesInFile <=  aNumLinesForDisplay ) {
      result = getAllOrFirst(aFile, null);
    }
    else {
      result = getOnlyFinalLines(aFile, numLinesInFile, aNumLinesForDisplay);
    }
    return result;
  }
  
  /** Return the number of lines in the given file.  */
  private int getNumLines(File aFile) throws DAOException {
    int result = 0;
    Scanner scanner = null;
    try {
      scanner = new Scanner(aFile);
    }
    catch(FileNotFoundException ex){
      throw new DAOException("Cannot find file " + Util.quote(aFile.getAbsolutePath()), ex);
    }
    while ( scanner.hasNextLine()) {
      scanner.nextLine();
      result++;
    }
    scanner.close();
    fLogger.fine("Number of lines in the file : " + result);
    return result;
  }
  
  /** Truncate file content to a gven number of lines at the end of the file.   */
  private String getOnlyFinalLines(File aFile, Integer aTotalLinesInFile, Integer aNumLinesForDisplay) throws DAOException {
    fLogger.fine("Truncating. Total lines in file : " + aTotalLinesInFile + " Num Lines desired for display :  " + aNumLinesForDisplay);
    StringBuilder result = new StringBuilder();
    assert(aTotalLinesInFile >= aNumLinesForDisplay);
    int start = aTotalLinesInFile - aNumLinesForDisplay;
    Scanner scanner = null;
    try {
      scanner = new Scanner(aFile);
    }
    catch(FileNotFoundException ex){
      throw new DAOException("Cannot find file " + Util.quote(aFile.getAbsolutePath()), ex);
    }
    int lineCount = 0;
    while ( scanner.hasNextLine()) {
      String line = scanner.nextLine();
      lineCount++;
      if ( lineCount >= start ) {
        result.append(line + NEW_LINE);
      }
    }
    scanner.close();
    return result.toString();
  }
  
  /** Return a FileFilter appropriate for the type of log.  */
  private FileFilter getFilter(LogFor aLogFor){
    FileFilter result = null; 
    final LogInfo dirInfo = fDirInfoDAO.fetch();
    if( LogFor.Server == aLogFor ) {
      result = acceptIfNameStartsWithSpecificText(dirInfo);
    }
    else if ( LogFor.Application == aLogFor ) {
      result = rejectEmptyFiles();
    }
    return result;  
  }

  /** Accept if non-zero size and name starts with specific text.  */
  private FileFilter acceptIfNameStartsWithSpecificText(LogInfo aDirInfo){
    final LogInfo dirInfo = aDirInfo;
    return new FileFilter() {
      public boolean accept(File aFile) {
        boolean nonZeroSize = (aFile.length() > 0);
        boolean fileNameStartsWith = aFile.getAbsolutePath().startsWith(dirInfo.getServerLoggingDirectory().getRawString() + dirInfo.getServerLogFileStartsWith());
        return nonZeroSize && fileNameStartsWith;
      }
    };
  }

  /** Accept only if non-zero size.  */
  private FileFilter rejectEmptyFiles(){
    return new FileFilter() {
      public boolean accept(File aFile) {
        return (aFile.length() > 0);
      }
    };
  }
  
  /**
  * Return the first {@link LoggerRecord} appearing in a file. 
  */
  private LoggerRecord firstRecordFor(File aLogFile) throws DAOException {
    //first N lines - will often cut off record at the end
    fLogger.fine("Getting first record for " + aLogFile.getName());
    String firstLines = getAllOrFirst(aLogFile, 100);
    LogParser parser = LogParserInstance.getFor(LogFor.Application);
    List<LoggerRecord> records = parser.parse(firstLines, ParsedCriteria.forDowntimeListing());
    fLogger.fine("Num logger records at start of file " + records.size());
    return records.get(0);
  }
}
