package hirondelle.web4jtools.logview.stats;

import hirondelle.web4j.action.ActionTemplateSearch;
import hirondelle.web4j.action.ResponsePage;
import hirondelle.web4j.model.AppException;
import hirondelle.web4j.request.RequestParameter;
import hirondelle.web4j.request.RequestParser;
import hirondelle.web4jtools.logview.parsedview.LoggerRecord;
import hirondelle.web4jtools.logview.parsedview.ParsedCriteria;
import hirondelle.web4jtools.logview.simpleview.LogFileDAO;

import java.io.File;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.logging.Level;

/**
* List stats for a log file, including a histogram of relative frequency of levels. 
*/
public final class LogStatsAction extends ActionTemplateSearch {

  public LogStatsAction(RequestParser aRequestParser){
    super(FORWARD, aRequestParser);
  }
  
  public static final RequestParameter LOG_FOR = RequestParameter.withRegexCheck("Log For", "(Application|Server)");
  
  /** Ensure user input can build a {@link ParsedCriteria}. */
  @Override protected void validateUserInput() throws AppException {
    //under normal operation, no chance of faulty user input - a single radio button, 
    //with a default selection
    fCriteria = ParsedCriteria.forStats(getParamUnsafe(LOG_FOR));
  }
  
  /** Show stats and histogram.  */
  @Override protected void listSearchResults() throws AppException {
    LogFileDAO dao = new LogFileDAO(getRequestParser().getRequest());
    File mostRecentLog = dao.getMostRecentLogFile(fCriteria.getLogFor());
    if ( mostRecentLog != null  ) {
      addToRequest("mostRecentLog", mostRecentLog);
      addToRequest("mostRecentLogDate", new Date(mostRecentLog.lastModified()));
      addToRequest("mostRecentLogSize", mostRecentLog.length());
      
      List<LoggerRecord> logRecords = dao.getParsedLogFile(mostRecentLog, fCriteria);
      addToRequest("totalRecords", logRecords.size());
      addToRequest("firstRecordDate", logRecords.get(0).getDate() );
      addToRequest("lastRecordDate", logRecords.get(logRecords.size()-1).getDate());
      calcHistogramBuckets(logRecords);
      addToRequest("buckets", fBuckets);
    }
    else {
      addError("No files detected in logging directory.");
    }
  }
  
  // PRIVATE //
  private ParsedCriteria fCriteria;
  private List<LevelBucket> fBuckets = new ArrayList<LevelBucket>();
  private static final ResponsePage FORWARD = new ResponsePage("Log Stats", "view.jsp", LogStatsAction.class);
  
  private void calcHistogramBuckets(List<LoggerRecord> aRecords) {
    buildLevelBuckets();
    for(LoggerRecord record : aRecords){
      addToAppropriateBucket(record);
    }
  }
  
  private void buildLevelBuckets() {
    addBucketFor(Level.SEVERE);
    addBucketFor(Level.WARNING);
    addBucketFor(Level.INFO);
    addBucketFor(Level.CONFIG);
    addBucketFor(Level.FINE);
    addBucketFor(Level.FINER);
    addBucketFor(Level.FINEST);
  }
  
  private void addBucketFor(Level aLevel) {
    fBuckets.add(new LevelBucket(aLevel));
  }
  
  private void addToAppropriateBucket(LoggerRecord aRecord){
    for (LevelBucket bucket : fBuckets){
      if ( aRecord.getLevel().equals(bucket.getLevel()) ) {
        bucket.incrementCount();
      }
    }
  }
}

