package hirondelle.web4jtools.metrics.base;

import hirondelle.web4j.model.ModelCtorException;
import hirondelle.web4j.security.SafeText;
import hirondelle.web4j.model.Check;
import hirondelle.web4j.model.ModelUtil;
import hirondelle.web4j.model.Validator;
import java.io.*;
import static hirondelle.web4j.util.Consts.FAILS;
import static hirondelle.web4j.util.Consts.FILE_SEPARATOR;
import hirondelle.web4j.util.Util;

/**
* Basic information on the target project. 
*/
public final class BaseInfo {
  
  /**
  * Full constructor.
  * 
  * @param aProjectName simple name or alias for the project, displayed as a reminder (required).
  * @param aBaseDirectory base or root directory for your project's source files.
  * @param aBaseURIForFetchingImages URI for running web app; when a simple image file 
  * name is concatenated with this URI, then the image will be fetched. Used for listing all image files in 
  * the target project.
  */ 
  public BaseInfo(SafeText aProjectName, SafeText aBaseDirectory, SafeText aBaseURIForFetchingImages) throws ModelCtorException {
    fProjecName = aProjectName;
    fBaseDirectory = aBaseDirectory;
    fBaseURIForFetchingImages = aBaseURIForFetchingImages;
    validateState();
  }
  
  public SafeText getProjectName() {  return fProjecName; }
  public SafeText getBaseDirectory() {  return fBaseDirectory; }
  public SafeText getBaseURIForFetchingImages() {  return fBaseURIForFetchingImages; }
  
  /** Intended for debugging only.  */
  @Override public String toString() {
    return ModelUtil.toStringFor(this);
  }

  @Override public boolean equals( Object aThat ) {
    Boolean result = ModelUtil.quickEquals(this, aThat);
    if ( result == null ){
      BaseInfo that = (BaseInfo) aThat;
      result = ModelUtil.equalsFor(this.getSignificantFields(), that.getSignificantFields());
    }
    return result;    
  }

  @Override public int hashCode() {
    if ( fHashCode == 0 ) {
      fHashCode = ModelUtil.hashCodeFor(getSignificantFields());
    }
    return fHashCode;
  }
  
  // PRIVATE //
  private SafeText fProjecName;
  private SafeText fBaseDirectory;
  private SafeText fBaseURIForFetchingImages;  
  private int fHashCode;
  
  private void validateState() throws ModelCtorException {
    ModelCtorException ex = new ModelCtorException();
    if( FAILS == Check.required(fProjecName) ) {
      ex.add("Project Name is required.");
    }
    if ( FAILS == Check.required(fBaseDirectory, validDirectory())) {
      ex.add("Base Directory is required: an existing, read-able directory, ending with a " + Util.quote(FILE_SEPARATOR) + ".");
    }
    if ( ex.isNotEmpty() ) throw ex;
  }
  
  private Validator validDirectory() {
    return new Validator() {
      public boolean isValid(Object aObject) {
        SafeText text = (SafeText)aObject;
        File file = new File(text.getRawString());
        return file.isDirectory() && file.exists() && file.canRead() && file.isAbsolute() && endsWithSeparator(text);
      }
    };
  }
  
  private boolean endsWithSeparator(SafeText aDir){
    return aDir.getRawString().endsWith(FILE_SEPARATOR);
  }
  
  private Object[] getSignificantFields(){
    return new Object[] {fProjecName, fBaseDirectory, fBaseURIForFetchingImages};
  }
}
