package hirondelle.web4jtools.metrics.base;

import hirondelle.web4j.action.ActionTemplateShowAndApply;
import hirondelle.web4j.action.ResponsePage;
import hirondelle.web4j.model.AppException;
import hirondelle.web4j.model.ModelCtorException;
import hirondelle.web4j.model.ModelFromRequest;
import hirondelle.web4j.request.RequestParameter;
import hirondelle.web4j.request.RequestParser;
import hirondelle.web4j.util.Stopwatch;
import hirondelle.web4j.util.Util;

import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;

/**
* Override base project information configured in <tt>web.xml</tt>, and show high level metrics to user.
* 
* <P>Scans the local file system under the given base directory.
* Various statistics and listings are extracted.
*/
public final class BaseInfoAction extends ActionTemplateShowAndApply {

  public BaseInfoAction(RequestParser aRequestParser) {
    super(FORWARD, REDIRECT, aRequestParser);
  }
  
  public static final RequestParameter BASE_DIRECTORY = RequestParameter.withLengthCheck("BaseDirectory");
  public static final RequestParameter PROJECT_NAME = RequestParameter.withLengthCheck("ProjectName");
  public static final RequestParameter BASE_URI_FOR_FETCHING_IMAGES = RequestParameter.withLengthCheck("BaseURIForFetchingImages");
  
  /** Show a form reflecting current settings.  */
  @Override protected void show() throws AppException {
     if( listFileInfo() == null ) {
       addMessage("Please 'Scan Source Code' to analyze your app's source code.");
     }
     addToRequest(ITEM_FOR_EDIT, fetchBaseInfo());
  }

  /** Validate changes to current settings.  */
  @Override protected void validateUserInput() throws AppException {
    ModelFromRequest builder = new ModelFromRequest(getRequestParser());
    try {
      fBaseInfo = builder.build(BaseInfo.class, PROJECT_NAME, BASE_DIRECTORY, BASE_URI_FOR_FETCHING_IMAGES);
    }
    catch (ModelCtorException ex) {
      addError(ex);
    }
  }
  
  /**
  * Save the changes to current settings, and recursively scan the files under  
  * the base directory.
  * 
  * <P>Creates an internal data structure to represent your application's source code.
  */
  @Override protected void apply() throws AppException {
    fLogger.fine("Saving current settings.");
    BaseInfoDAO dao  = new BaseInfoDAO(getRequestParser().getRequest());
    dao.save(fBaseInfo);
    
    fLogger.fine("Scanning source files...");
    Stopwatch stopwatch = new Stopwatch();
    stopwatch.start();
    
    File baseDirectory = new File(fBaseInfo.getBaseDirectory().getRawString());
    scan(baseDirectory);
    
    addToSession(FILE_INFO_MAP_KEY, fAllFiles);
    addToSession(FILE_INFO_LIST_KEY, asSortedList(fAllFiles));
    
    addToSession("numFiles", fNumFiles);
    addToSession("totalSize", fTotalSize);
    addToSession("numSourceFiles", fNumSourceFiles);
    addToSession("numImageFiles", fNumImageFiles);
    addToSession("numMarkupFiles", fNumMarkupFiles);
    addToSession("numJavaFiles", fNumJavaFiles);
    
    addToSession("sourceFileExtensions", FileInfo.getSourceFileExtensions());
    addToSession("imageFileExtensions", FileInfo.getImageFileExtensions());
    addToSession("markupFileExtensions", FileInfo.getMarkupFileExtensions());
    addToSession("ignorableFiles", FileInfo.getIgnorableFiles());
    
    stopwatch.stop();
    fLogger.fine("Finished scanning source files. Time needed for scan : " + stopwatch + " msec.'");
  }
  
  /**
  * Key for <tt>Map</tt> stored in session scope. This data structure is created/replaced when 
  * the user presses the 'Scan Source Code' button.
  * 
  * <P>This Map style retains directory information.
  */
  public static final String FILE_INFO_MAP_KEY = "fileInfoMap";
  
  /**
  * Key for <tt>List</tt> stored in session scope. This data structure is created/replaced when 
  * the user presses the 'Scan Source Code' button.
  * 
  * <P>This List style removes directory information, and is useful for file listings. 
  * When the user changes the sorting, the re-sorted list should overwrite the object
  * in session scope. That way, the latest sort used by the user will be retained.
  */
  public static final String FILE_INFO_LIST_KEY = "fileInfoList";

  // PRIVATE //
  private BaseInfo fBaseInfo;
  private Map<File, List<FileInfo>> fAllFiles = new LinkedHashMap<File, List<FileInfo>>();
  private long fNumFiles;
  private long fTotalSize;
  private long fNumSourceFiles;
  private long fNumJavaFiles;
  private long fNumImageFiles;
  private long fNumMarkupFiles;
  
  private static final ResponsePage FORWARD = new ResponsePage("Source Code Base Info", "view.jsp", BaseInfoAction.class);
  private static final ResponsePage REDIRECT = new ResponsePage("BaseInfoAction.do?Operation=Show");
  private static final Logger fLogger = Util.getLogger(BaseInfoAction.class);

  private Map<File, List<FileInfo>> listFileInfo() {
    Map<File, List<FileInfo>> result = (Map<File, List<FileInfo>>)getFromSession(FILE_INFO_MAP_KEY);
    return result;
  }
  
  private BaseInfo fetchBaseInfo(){
    BaseInfoDAO dao = new BaseInfoDAO(getRequestParser().getRequest());
    return dao.fetch();
  }

  /**
  * Recursively scan the base dir. When each file is encountered, it is examined.
  * A FileInfo thing is created, and it is added to allFiles.
  * Along the way, the various counters are incremented, in order to avoid 
  * possible performance problems with large numbers of files.
  */
  private void scan(File aDirectory) {
    assert (aDirectory.isDirectory());
    List<FileInfo> filesOnly = new ArrayList<FileInfo>();
    List<File> allInDirectory = Arrays.asList(aDirectory.listFiles());
    for(File file : allInDirectory){
      if(! file.isDirectory() ) {
        addFile(filesOnly, file);
      }
    }
    fAllFiles.put(aDirectory, filesOnly);
    for( File dir : allInDirectory ) {
      if ( dir.isDirectory() ) {
        scan(dir); //recursive call!!
      }
    }
  }

  private void addFile(List<FileInfo> aFilesOnly, File aFile) {
    if ( FileInfo.isIgnorable(aFile)) {
      //do nothing
    }
    else{
      FileInfo fileInfo = new FileInfo(aFile);
      aFilesOnly.add(fileInfo);
      fNumFiles = fNumFiles + 1; 
      fTotalSize = fTotalSize + aFile.length();
      if ( fileInfo.isSourceFile() ) {
        ++fNumSourceFiles;
      }
      if ( fileInfo.isJavaSourceFile() ) {
        ++fNumJavaFiles;
      }
      if ( fileInfo.isImageFile() ) {
        ++fNumImageFiles;
      }
      if ( fileInfo.isMarkupFile() ) {
        ++fNumMarkupFiles;
      }
    }
  }
  
  private List<FileInfo> asSortedList(Map<File, List<FileInfo>> aAllFiles){
    List<FileInfo> result = new ArrayList<FileInfo>();
    Set<File> dirs =  aAllFiles.keySet();
    for ( File dir : dirs ) {
      List<FileInfo> files = aAllFiles.get(dir);
      result.addAll(files);
    }
    Collections.sort(result, byTotalSize());
    return result;
  }
  
  private Comparator<FileInfo> byTotalSize() {
    return new Comparator<FileInfo>() {
      public int compare(FileInfo aThis, FileInfo aThat) {
        return aThis.getSize().compareTo(aThat.getSize());
      }
    };
  }
}
