package hirondelle.web4jtools.metrics.images;

import static hirondelle.web4j.util.Consts.NOT_FOUND;
import hirondelle.web4j.action.ResponsePage;
import hirondelle.web4j.request.RequestParameter;
import hirondelle.web4j.request.RequestParser;
import hirondelle.web4j.util.Util;
import hirondelle.web4jtools.metrics.base.FileInfo;
import hirondelle.web4jtools.metrics.base.MetricsAction;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
;

/**
* List all images in the target project, along with all references to each image occurring in 
* markup files.
* Any image having 0 references are candidates for removal from the project. 
*/
public class ImageReferencesAction extends MetricsAction {

  /** Full constructor. */
  public ImageReferencesAction(RequestParser aRequestParser){
    super(FORWARD, aRequestParser);
  }
  
  /** The column of the listing to sort on. See the <tt>sortLinks.tag</tt> file. */
  public static final RequestParameter SORT_ON = RequestParameter.withRegexCheck("SortOn", "(Name|NumReferences|Bytes)");
  /** Ascending or descending order. See the <tt>sortLinks.tag</tt> file. */
  public static final RequestParameter ORDER = RequestParameter.withRegexCheck("Order", "(ASC|DESC)");
  
  /** 
  * Show a sortable listing of images and the files that reference them.
  * The default sort is by the number of references.
  */
  @Override  protected void calculateMetric() {
    List<ImageReference> imgRefs = new ArrayList<ImageReference>();
    List<FileInfo> imageFiles = filterImageFiles();
    List<FileInfo> markupFiles = filterMarkupFiles();
    initializeImageReferences(imgRefs, imageFiles);

    for(FileInfo markupFile : markupFiles){
      String markupFileContent = markupFile.getContent();
      for (FileInfo imageFile : imageFiles) {
        if( hasImageReference(markupFileContent, imageFile)){
          addImageReference(markupFile, imageFile, imgRefs);
        }
      }
    }
    sort(imgRefs);
    addToRequest("imgReferences", imgRefs);
  }
  
  
  // PRIVATE //
  private static final ResponsePage FORWARD = new ResponsePage("Images and Their References", "view.jsp", ImageReferencesAction.class);
  
  private List<FileInfo> filterImageFiles() {
    List<FileInfo> result = new ArrayList<FileInfo>();
    for(FileInfo file : fFileList){
      if ( file.isImageFile() ) {
        result.add(file);
      }
    }
    return result;
  }
  
  private List<FileInfo> filterMarkupFiles() {
    List<FileInfo> result = new ArrayList<FileInfo>();
    for(FileInfo file : fFileList){
      if ( file.isMarkupFile() ) {
        result.add(file);
      }
    }
    return result;
  }
  
  private void  initializeImageReferences(List<ImageReference> aImgRefs, List<FileInfo> aImageFiles){
    for(FileInfo imageFile : aImageFiles)  {
      aImgRefs.add(new ImageReference(imageFile));
    }
  }
  
  /**
  * This check is a bit simplistic, since it does not use path information at all. 
  * It only checks if the file contains the simple image file name, anywhere in its 
  * body, without using the full path of the image file. 
  * 
  * <P>This style at least allows for referencing image files in many different ways.
  */
  private boolean hasImageReference(String aMarkupFileContent, FileInfo aImageFile){
    return aMarkupFileContent.indexOf(aImageFile.getSimpleName().getRawString()) != NOT_FOUND;
  }

  private void addImageReference(FileInfo aMarkupFile, FileInfo aImageFile, List<ImageReference> aImgRefs) {
    ImageReference imageRef = null;
    for (ImageReference item : aImgRefs){
      if (item.getFileInfo().equals(aImageFile)) {
        imageRef = item;
        break;
      }
    }
    if ( imageRef == null ) {
      throw new AssertionError("Cannot find image file as expected.");
    }
    imageRef.addReferenceFrom(aMarkupFile);
  }
  
  private void sort(List<ImageReference> aImgRefs) {
    String sortOn = getParamUnsafe(SORT_ON);
    String order = getParamUnsafe(ORDER);
    Collections.sort(aImgRefs, getComparator(sortOn));
    if( "DESC".equals(order) || ! Util.textHasContent(order)) {
      Collections.reverse(aImgRefs);
    }
  }
  
  private Comparator<ImageReference> getComparator(String aSortOn){
    Comparator<ImageReference> result = null;
    if( "Name".equals(aSortOn)  ) {
      result = new Comparator<ImageReference>() {
        public int compare(ImageReference aThis, ImageReference aThat) {
          String thisName = aThis.getFileInfo().getSimpleName().getRawString();
          String thatName = aThat.getFileInfo().getSimpleName().getRawString();
          return thisName.compareTo(thatName);
        };
      };
    }
    if( "Bytes".equals(aSortOn)  ) {
      result = new Comparator<ImageReference>() {
        public int compare(ImageReference aThis, ImageReference aThat) {
          Long thisSize = aThis.getFileInfo().getSize();
          Long thatSize = aThat.getFileInfo().getSize();
          return thisSize.compareTo(thatSize);
        };
      };
    }
    //default sorting by num references
    else if ( "NumReferences".equals(aSortOn) || ! Util.textHasContent(aSortOn) ) {
      result = new Comparator<ImageReference>() {
        public int compare(ImageReference aThis, ImageReference aThat) {
          return aThis.getNumReferences().compareTo(aThat.getNumReferences());
        };
      };
    }
    else {
      throw new AssertionError("Unknown sort column : " + Util.quote(aSortOn));
    }
    return result;
  }
}
