package hirondelle.web4jtools.metrics.tabs;

import static hirondelle.web4j.util.Consts.EMPTY_STRING;
import static hirondelle.web4j.util.Consts.NEW_LINE;
import static hirondelle.web4j.util.Consts.SPACE;
import hirondelle.web4j.action.ResponsePage;
import hirondelle.web4j.request.RequestParser;
import hirondelle.web4j.util.Util;
import hirondelle.web4jtools.metrics.base.FileInfo;
import hirondelle.web4jtools.metrics.base.MetricsAction;
import hirondelle.web4jtools.util.Ensure;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.Iterator;
import java.util.Scanner;
import java.util.logging.Logger;

import javax.servlet.ServletConfig;

/**
* Replace all tab characters in source code with spaces.
* The number of spaces used is controlled by the 
* <tt>NumSpacesForTabs</tt> setting in <tt>web.xml</tt>.
* 
* <P>Since this action can potentially make a large number of <em>irreversible</em>
* edits to your source code, it is prudent to make a back up copy of your source before 
* executing. 
*/
public class TabsRemoveAction extends MetricsAction {

  /** Read in config from web.xml.  Called only upon startup. */
  public static void readConfig(ServletConfig aConfig){
    fNUM_SPACES_FOR_TABS = aConfig.getInitParameter(NUM_SPACES_FOR_TABS);
    Ensure.isPresentInWebXml(NUM_SPACES_FOR_TABS, fNUM_SPACES_FOR_TABS);
    if( Integer.valueOf(fNUM_SPACES_FOR_TABS) < 1 ) {
      throw new IllegalArgumentException(NUM_SPACES_FOR_TABS + " setting in web.xml must be 1 or more. Value : " + Util.quote(fNUM_SPACES_FOR_TABS));
    }
    fREPLACEMENT_FOR_TAB = getReplacementForTabs();
    fENCODING = aConfig.getInitParameter(ENCODING);
    Ensure.isPresentInWebXml(ENCODING, fENCODING);
  }
  
  /** Full constructor.  */
  public TabsRemoveAction(RequestParser aRequestParser) {
    super(FORWARD, aRequestParser);
  }

  /** 
  * For all source code files, replace tab characters with spaces. 
  */
  @Override protected void calculateMetric() {
    int tabsCount = 0;
    Iterator<FileInfo> iter = fFileList.iterator();
    while ( iter.hasNext() ){
      FileInfo file = iter.next();
      if ( file.getNumTabs() > 0 ) {
        fLogger.fine("Replacing tabs in " + file.getName().getRawString());
        tabsCount = tabsCount + replaceTabsAndUpdateFileInfo(file);
      }
    }
    fLogger.fine("Total number of tab characters replaced in source files: " + tabsCount);
    addMessage("Please re-scan your source code (Summary page) to verify no TAB chars remain.");
    setResponsePage(REDIRECT_TO_LISTING);
  }
  
  // PRIVATE  //
  private static final String NUM_SPACES_FOR_TABS = "NumSpacesForTabs";
  private static String fNUM_SPACES_FOR_TABS;
  private static String fREPLACEMENT_FOR_TAB;
  private static final String ENCODING = "CharacterSetForGeneratedFiles";
  private static String fENCODING;
  
  private static final ResponsePage FORWARD = new ResponsePage("Files With TAB Chars", "view.jsp", TabsRemoveAction.class);
  private static final ResponsePage REDIRECT_TO_LISTING = new ResponsePage("TabsListAction.do");
  
  private static final Logger fLogger = Util.getLogger(TabsRemoveAction.class);
  
  private int replaceTabsAndUpdateFileInfo(FileInfo aFile) {
    assert(aFile.getNumTabs() > 0);
    int result = aFile.getNumTabs();
    File file = new File(aFile.getName().getRawString());
    String fileContent = getContentOf(file);
    String updatedContent = replaceTabs(fileContent);
    save(updatedContent, file);
    return result;
  }
  
  private String getContentOf(File aFile){
    StringBuilder result = new StringBuilder();
    try { 
      Scanner scanner = new Scanner(aFile);
      while ( scanner.hasNextLine() ) {
        result.append(scanner.nextLine() + NEW_LINE);
      }
      scanner.close();
    }
    catch(FileNotFoundException ex){
      fLogger.severe("Cannot open file named " + aFile.getName());
    }
    return result.toString();
  }
  
  private static String getReplacementForTabs(){
    String result = EMPTY_STRING;
    Integer numSpaces = Integer.valueOf(fNUM_SPACES_FOR_TABS);
    for (int idx=1; idx <= numSpaces; ++idx){
      result = result + SPACE;
    }
    fLogger.fine("Replacement for tabs : " + Util.quote(result));
    return result;
  }
  
  private String replaceTabs(String aFileContent){
    return aFileContent.replace("\t", fREPLACEMENT_FOR_TAB);
  }
  
  /** Write the new contents to the file, using encoding specified in web.xml.  */
  private void  save(String aUpdatedContent, File aFile){
    fLogger.fine("Updating contents of " + aFile.getAbsolutePath());
    Writer output = null;
    try {
      FileOutputStream fos = new FileOutputStream(aFile.getAbsolutePath());
      OutputStreamWriter out = new OutputStreamWriter(fos, fENCODING); 
      output = new BufferedWriter(out);
      output.write( aUpdatedContent );
    }
    catch (IOException ex){
      fLogger.severe("Replacing tags, but cannot write to file named : " + Util.quote(aFile.getAbsolutePath()));
    }
    finally {
      try {
        if (output != null) output.close();
      }
      catch (IOException ex) {
        fLogger.severe("Replacing tags, but cannot close file named : " + Util.quote(aFile.getAbsolutePath()));
      }
    }    
  }
}
