package hirondelle.web4jtools.taglet;

import java.util.*;
import com.sun.tools.doclets.Taglet;    
import com.sun.javadoc.Tag;               
import static hirondelle.web4j.util.Consts.NEW_LINE;
import hirondelle.web4j.util.Util;

/**
* Taglet for linking to <tt>.sql</tt> files.
*
* <P>The <tt>.sql</tt> files need to be copied into the javadoc tree. 
* The typical use case is to place the <tt>.sql</tt> file in the same 
* directory as the class that references it. 
* 
* <P>Example 1 :<br>
* <tt>&ampsql statements.sql</tt>
*  
* <P>Example 2 :<br>
* <tt>&ampsql ../blah/statements.sql</tt>
* 
*  <P>This references a file in a nearby directory.
*  
* <P>Example 3 :<br>
* <tt>&ampsql statements.sql ADD_BLAH, CHANGE_BLAH</tt>
* <P>This example references two specific SQL statements in the <tt>statements.sql</tt> file. 
* That is, it references only a subset of the statements in the file.
*/
public final class SqlTaglet implements Taglet {
  
  /** Register this taglet with the javadoc tool, under the taglet name 'sql'.  */
  public static void register(Map aTagletMap){
    if ( aTagletMap.get(TAGLET_NAME) != null ) {
      aTagletMap.remove(TAGLET_NAME);
    }
    aTagletMap.put(TAGLET_NAME, new SqlTaglet());
  }
  
  /** Return 'sql'. */
  public String getName() {
    return TAGLET_NAME;
  }
  
  /** Returns <tt>false</tt>.  */
  public boolean isInlineTag() {
    return false;
  }
  
  /** Returns <tt>false</tt>.  */
  public boolean inOverview() {
    return false;
  }
  
  /** Returns <tt>false</tt>.  */
  public boolean inPackage() {
    return false;
  }
  
  /** Returns <tt>true</tt>.  */
  public boolean inType() {
    return true;
  }
  
  /** Returns <tt>false</tt>.  */
  public boolean inConstructor() {
    return false;
  }
  
  /** Returns <tt>false</tt>.  */
  public boolean inMethod() {
    return false;
  }
  
  /** Returns <tt>false</tt>.  */
  public boolean inField() {
    return false;
  }
  
  /** Render a series of <tt>&amp;sql</tt> items.  */
  public String toString(Tag[] aTags) {
    StringBuilder result = new StringBuilder();
    if ( aTags.length != 0) {
      result.append(TAGLET_HEADER);
      result.append(NEW_LINE);
      result.append("<ul>" + NEW_LINE);
      for(Tag tag: aTags){
        if( Util.textHasContent(tag.text()) ) {
          result.append("<li>" + getLink(tag) + NEW_LINE);
        }
      }
      result.append("</ul>" + NEW_LINE);
    }
    return result.toString();
  }

  /** Render a single <tt>&amp;sql</tt> item.  */
  public String toString(Tag aTag) {
    String result = "";
    if( Util.textHasContent(aTag.text()) )  {
      result = TAGLET_HEADER + getLink(aTag) + NEW_LINE;
    }
    return result;
  }
  
  // PRIVATE //
  private static final String TAGLET_NAME = "sql";
  private static final String TAGLET_HEADER = "<b>SQL Statements :</b> ";
  
  private String getLink(Tag aTag){
    ParsedTag parsedTag = new ParsedTag(aTag);
    String result = "<a href='" + parsedTag.getFileName() + "'><tt>" + parsedTag.getFileName() + "</tt></a>";
    if ( ! parsedTag.getOperations().isEmpty() ){
      result = result + getOperation(parsedTag);
      for (String operation : parsedTag.getOperations() ){
        result = result + " <tt>" + operation + "</tt>";
      }
    }
    return result;
  }
  
  /** Return the list of operations, if any. */
  private String getOperation(ParsedTag aParsedTag) {
    String result = "";
    if ( aParsedTag.getOperations().size() == 1 ){
      result = result + " - Single Operation";
    }
    else {
      result = result + " -  Operations (" + aParsedTag.getOperations().size() + ")";
    }
    result = result + ": ";
    return result;
  }
  
  /**
  * Parse a tag's text content into two parts: a link/file name, and an optional list of operations.
  * The operations are simply the names attached to SQL statements, in .sql files.
  */
  private static final class ParsedTag {
    ParsedTag(Tag aTag){
      parseTag(aTag.text());
    }
    String getFileName() { return fFileName; }
    List<String> getOperations() {  return fOperations;   }
    private void parseTag(String aRawTag){
      StringTokenizer parser = new StringTokenizer(aRawTag);
      fFileName = parser.nextToken();
      while ( parser.hasMoreTokens() ) {
        fOperations.add(parser.nextToken());
      }
    }
    private String fFileName;
    private List<String> fOperations = new ArrayList<String>();
  }
}
