package hirondelle.web4jtools.webmaster.diagnostics; 

import hirondelle.web4j.Controller;
import hirondelle.web4j.action.ActionImpl;
import hirondelle.web4j.action.ResponsePage;
import hirondelle.web4j.database.DAOException;
import hirondelle.web4j.model.AppException;
import hirondelle.web4j.request.RequestParser;
import hirondelle.web4j.util.Stopwatch;
import hirondelle.web4j.util.Util;

import java.io.IOException;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.jar.Attributes;
import java.util.jar.JarInputStream;
import java.util.jar.Manifest;
import java.util.logging.Level;
import java.util.logging.LogManager;
import java.util.logging.Logger;

import javax.servlet.ServletContext;
import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

/**
* Show an extensive listing of diagnostic information, useful for solving problems.
*/
public final class ShowDiagnostics extends ActionImpl {
  
  /** Constructor. */
  public ShowDiagnostics(RequestParser aRequestParser){
    super(FORWARD, aRequestParser);
  }
  
  /** 
  * Retrieve diagnostic information and display it to the user.
  * 
  * <P>The diagnostic information includes :
  * <ul>
  * <li>up time
  * <li>logging configuration
  * <li>jar versions
  * <li>system properties
  * <li>context init params
  * <li>server information
  * <li>application scope items
  * <li>session scope items
  * <li>request information
  * <li>request headers
  * <li>request cookies
  * <li>response encoding
  * </ul> 
  */
  public ResponsePage execute() throws AppException {
    placeDiagnosticDataInScope(getRequestParser().getRequest(), getRequestParser().getResponse());
    return getResponsePage();
  }
  
  // PRIVATE //
  private static final ResponsePage FORWARD = new ResponsePage(
    "Diagnostics", "view.jsp", ShowDiagnostics.class
  );
  private static final String SPECIFICATION_TITLE = "Specification-Title";
  private static final String SPECIFICATION_VERSION = "Specification-Version";
  private static final String UNSPECIFIED = "Unspecified in Manifest";
  private static final long MILLISECONDS_PER_DAY = 1000*60*60*24L;
  private static final Logger fLogger = Util.getLogger(ShowDiagnostics.class);
  
  private void placeDiagnosticDataInScope(HttpServletRequest aRequest, HttpServletResponse aResponse) throws DAOException {
    Stopwatch stopwatch = new Stopwatch();
    stopwatch.start();
    
    fLogger.fine("Adding system properties.");
    addToRequest("systemProperties", sortMap(System.getProperties()));
    fLogger.fine("Adding Context init-params.");
    addToRequest("contextInitParams", sortMap(getContextInitParams(aRequest)));
    fLogger.fine("Adding application scope items.");
    addToRequest("appScopeItems", getAppScope(aRequest));
    fLogger.fine("Adding session scope items.");
    addToRequest("sessionScopeItems", getSessionScope(aRequest));
    fLogger.fine("Adding container/servlet info.");
    addToRequest("containerServletInfo", getContainerServletInfo(aRequest));
    fLogger.fine("Adding request info.");
    addToRequest("requestInfo", getRequestInfo(aRequest));
    fLogger.fine("Adding loggers.");
    addToRequest("loggers", getLoggers());
    fLogger.fine("Adding JAR versions.");
    addToRequest("jarVersions", getJarVersions(aRequest));
    fLogger.fine("Adding uptime.");
    addToRequest("uptime", getUptime(aRequest));
    fLogger.fine("Adding request headers.");
    addToRequest("headers", getHeaders(aRequest));
    addToRequest("responseEncoding", getResponseEncoding(aResponse));
    fLogger.fine("Adding cookies.");
    addToRequest("cookies", getCookies(aRequest));
    
    fLogger.fine("Finished retrieving data.");
    stopwatch.stop();
    addToRequest("stopwatch", stopwatch.toString());
  }
  
  private ServletContext getContext(HttpServletRequest aRequest){
    return aRequest.getSession().getServletContext();
  }
  
  private Map sortMap(Map aInput){
    Map result = new TreeMap(String.CASE_INSENSITIVE_ORDER);
    result.putAll(aInput);
    return result;
  }
  
  private Map<Object, Object> getAppScope(HttpServletRequest aRequest){
    Map<Object, Object> result = new HashMap<Object, Object>();
    Enumeration keys = getContext(aRequest).getAttributeNames();
    while ( keys.hasMoreElements() ) {
      Object key = keys.nextElement();
      Object value = getContext(aRequest).getAttribute(key.toString());
      result.put(key, value);
    }
    return sortMap(result);
  }
  
  
  private Map<Object, Object> getSessionScope(HttpServletRequest aRequest){
    Map<Object, Object> result = new HashMap<Object, Object>();
    Enumeration keys = aRequest.getSession(true).getAttributeNames();
    while ( keys.hasMoreElements() ) {
      Object key = keys.nextElement();
      Object value = aRequest.getSession(true).getAttribute(key.toString());
      result.put(key, value);
    }
    return sortMap(result);
  }
  
  private Map<String, String> getContainerServletInfo(HttpServletRequest aRequest){
    Map<String, String> result = new LinkedHashMap<String, String>();
    ServletContext context = getContext(aRequest);
    result.put("Host Name", aRequest.getServerName());
    result.put("Operating System", System.getProperty("os.arch") + " " +  System.getProperty("os.name") + " " + System.getProperty("os.version"));
    result.put("Container", context.getServerInfo());
    result.put("Controller Name", context.getClass().getName());
    result.put("Home URL", getHome(aRequest));
    return result;
  }
  
  private Map<String, Object> getRequestInfo(HttpServletRequest aRequest){
    Map<String, Object> result = new HashMap<String, Object>();
    result.put("Server Port", new Integer(aRequest.getServerPort()));
    result.put("Client IP", aRequest.getRemoteAddr());
    result.put("Character Encoding", aRequest.getCharacterEncoding());
    result.put("Protocol", aRequest.getProtocol());
    result.put("Server Name", aRequest.getServerName());
    result.put("Content Length", new Integer(aRequest.getContentLength()));
    return sortMap(result);
  }

  private String getHome(HttpServletRequest aRequest) {
    String url =  aRequest.getRequestURL().toString();
    String contextPath = aRequest.getContextPath();
    fLogger.fine( url );
    fLogger.fine( contextPath );
    int endIndex = url.indexOf(contextPath) + contextPath.length();
    return url.substring(0, endIndex);
  }
  
  private Map<String, Level> getLoggers() {
    Map<String, Level> result = new HashMap<String, Level>();
    LogManager logManager = LogManager.getLogManager();
    Enumeration loggerNames = logManager.getLoggerNames();
    while ( loggerNames.hasMoreElements() ){
      String loggerName = (String)loggerNames.nextElement();
      Logger logger = logManager.getLogger(loggerName);
      result.put(loggerName, logger.getLevel());
    }
    return sortMap(result);
  }
  
  /**
  * Get all .jar files in <tt>/WEB-INF/lib/</tt> directory, scan the manifest for its 
  * <tt>Specification-Version</tt>, and report it along with the file name. 
  * 
  * <P>Some jars will not have a value for <tt>Specification-Version</tt>. These will be 
  * reported as {@link #UNSPECIFIED}.
  */  
  private Map<String, String> getJarVersions(HttpServletRequest aRequest){
    Map<String, String> result = new HashMap<String, String>();
    ServletContext context = getContext(aRequest); 
    Set paths = context.getResourcePaths("/WEB-INF/lib/");
    fLogger.fine(paths.toString());
    Iterator iter = paths.iterator();
    while ( iter.hasNext() ) {
      String jarFileName = iter.next().toString();
      JarInputStream jarStream = null;
      try {
        jarStream = new JarInputStream(context.getResourceAsStream(jarFileName));
      }
      catch (IOException ex){
        fLogger.severe("Cannot open jar file (to fetch Specification-Version from the jar Manifest).");
      }
      result.put(jarFileName, fetchSpecNamesAndVersions(jarStream));
    }
    return sortMap(result);
  }

  /** 
  * A Jar can have more than one entry for spec name and version. 
  * For example, the servlet jar implements both servlet spec and jsp spec.
  * 
  * <P>Search for all attributes named Specification-Title and Specification-Version, 
  * and simply return them in the order they appear.
  * 
  * <P>If none of these appear, simply return an empty String.
  */
  private String fetchSpecNamesAndVersions(JarInputStream aJarStream) {
    StringBuffer result = new StringBuffer();
    Manifest manifest = aJarStream.getManifest();
    if ( manifest != null ){
      Attributes attrs = (Attributes)manifest.getMainAttributes();
      for (Iterator iter = attrs.keySet().iterator(); iter.hasNext(); ) {
        Attributes.Name attrName = (Attributes.Name)iter.next();
        addSpecAttrIfPresent(SPECIFICATION_TITLE, result, attrs, attrName);
        addSpecAttrIfPresent(SPECIFICATION_VERSION, result, attrs, attrName);
      }
      fLogger.fine("Specification-Version: " + result);
    }
    else {
      fLogger.fine("No manifest.");
    }
    return result.toString();
  }

  private void addSpecAttrIfPresent(String aName, StringBuffer aResult, Attributes aAttrs, Attributes.Name aAttrName) {
    if ( aName.equalsIgnoreCase(aAttrName.toString()) ) {
      if ( Util.textHasContent(aAttrs.getValue(aAttrName)) ){
        aResult.append(aAttrs.getValue(aAttrName) + " ");
      }
    }
  }
  
  private Float getUptime(HttpServletRequest aRequest){
    java.util.Date now = new java.util.Date();
    java.util.Date startup = (java.util.Date)aRequest.getSession(true).getServletContext().getAttribute(Controller.START_TIME);
    float msecs = now.getTime() - startup.getTime();
    return new Float(msecs/MILLISECONDS_PER_DAY);
  }
  
  private Map<String, String> getHeaders(HttpServletRequest aRequest){
    Map<String, String> result = new HashMap<String, String>();
    Enumeration items = aRequest.getHeaderNames();
    while( items.hasMoreElements() ){
      String name = (String)items.nextElement();
      String value = aRequest.getHeader(name);
      result.put(name, value);
    }
    return sortMap(result);
  }

  private Map<String, String> getResponseEncoding(HttpServletResponse aResponse){
    Map<String, String> result = new HashMap<String, String>();
    result.put("Character Encoding", aResponse.getCharacterEncoding());
    return sortMap(result);
  }
  
  private Map<String, String> getCookies(HttpServletRequest aRequest){
    Map<String, String> result = new HashMap<String, String>();
    Cookie[] cookies = aRequest.getCookies();
    if ( cookies != null ){
      List cookieList = Arrays.asList(cookies);
      Iterator iter = cookieList.iterator();
      while ( iter.hasNext() ) {
        Cookie cookie = (Cookie)iter.next();
        result.put(cookie.getName(), "Value:" + cookie.getValue() + ", Comment:" + cookie.getComment() + ", Domain:" + cookie.getDomain() + ", Max-Age:" + cookie.getMaxAge() + ", Path:" + cookie.getPath() + ", Spec- Version:" + cookie.getVersion());
      }
      result = sortMap(result);
    }
    return result;
  }
  
  private Map<String, Object> getContextInitParams(HttpServletRequest aRequest){
    Map<String, Object> result = new HashMap<String, Object>();
    Enumeration initParams = aRequest.getSession().getServletContext().getInitParameterNames();
    while(initParams.hasMoreElements()) {
      String key = (String)initParams.nextElement();
      Object value = aRequest.getSession().getServletContext().getInitParameter(key);
      result.put(key, value);
    }
    return result;
  }
}
